/*
	bitboard.h	Game Board Operations for BitBoard
	Copyright (c) 2004 Kriang Lerdsuwanakij
	email:		lerdsuwa@users.sourceforge.net

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef BITBOARD_H
#define BITBOARD_H

#include "config.h"

#include "board.h"

struct bit_board_type {
	int	black[2];	// Row 1-4, 5-8
	int	white[2];	// Row 1-4, 5-8
};
extern bit_board_type bit_board_begin;

inline int xy_to_byte_pos(int x, int y)
{
	return y / 4;
}

inline int xy_to_bit_pos(int x, int y)
{
	return (y & 3) * 8 + x;
}

inline int pos_to_byte_pos(int pos)
{
	return pos / 32;
}

inline int pos_to_bit_pos(int pos)
{
	return pos % 32;
}

int count_bits(int x);
struct bit_board_info {
	bit_board_type	board;
	int		hash;
	char		move;
	bit_board_info();
	bit_board_info(const bit_board_type *board_init);
};

bool operator==(const bit_board_info &board1, const bit_board_info &board2);

inline bool is_empty(const bit_board_info *board, int x, int y)
{
	int mask = 0x80000000 >> xy_to_bit_pos(x, y);
	int byte_pos = xy_to_byte_pos(x, y);
	return ((board->board.black[byte_pos]
		 | board->board.white[byte_pos]) & mask) == 0;
}

inline bool is_empty(const bit_board_info *board, int pos)
{
	int mask = 0x80000000 >> pos_to_bit_pos(pos);
	int byte_pos = pos_to_byte_pos(pos);
	return ((board->board.black[byte_pos]
		 | board->board.white[byte_pos]) & mask) == 0;
}

void	copy_board(bit_board_info *dest, const bit_board_info *src);

inline void copy_board(bit_board_info &dest, const bit_board_info &src)
{
	copy_board(&dest, &src);
}

inline void copy_board(bit_board_info *dest, const bit_board_type *src)
{
	bit_board_info temp_board(src);
	copy_board(dest, &temp_board);
}

inline void copy_board(bit_board_info &dest, const bit_board_type *src)
{
	copy_board(&dest, src);
}

inline int	board_black_score(const bit_board_info *board)
{
	return count_bits(board->board.black[0])+count_bits(board->board.black[1]);
}
inline int	board_white_score(const bit_board_info *board)
{
	return count_bits(board->board.white[0])+count_bits(board->board.white[1]);
}
inline int	board_diff_score(const bit_board_info *board)
{
	return board_black_score(board) - board_white_score(board);
}
#endif /* BITBOARD_H */
