/*
	game.h		Game States
	Copyright (c) 2004 Kriang Lerdsuwanakij
	email:		lerdsuwa@users.sourceforge.net

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <pthread.h>

#include <pwd.h>
#include <sys/types.h>
#include <time.h>

#include CXX__HEADER_exception
#include CXX__HEADER_stdexcept
#include CXX__HEADER_fstream
#include CXX__HEADER_sstream
#include CXX__HEADER_iomanip
#include <string>
#include <deque>
#include <vector>

USING_NAMESPACE_STD;

#include "board.h"

/*************************************************************************
	AI variables
*************************************************************************/

extern pthread_t	ai_thread;
extern pthread_attr_t	ai_thread_attr;
extern pthread_mutex_t	ai_lock;
extern volatile int	ai_pos;			// Lock required for access
extern gint		ai_timeout_id;		// Lock required for access
extern bool		ai_tampered;

/*************************************************************************
	Game states
*************************************************************************/

extern board_info *board_ptr;
extern board_info *init_board_ptr;

extern int player;
extern bool wait_player;
extern bool game_play;
extern bool random_game;

// Opening information included
extern board_info	board_history[NUM_MOVE+1];
extern int		player_history[NUM_MOVE+1];
extern int		move_history[NUM_MOVE+1];
extern int		num_history;
extern int		min_num_history;
extern int		max_num_history;

extern deque<int>	move_queue;

extern string		home_dir;

				// Useful hint to manage move lists
enum update_state_type {
	UPDATE_ALL,		// Game changed
	UPDATE_BACKWARD,	// Moves removed
	UPDATE_FORWARD,		// Moves added
	UPDATE_NONE		// No moves, repaint board/status
};
typedef void (*update_func)(update_state_type);
extern vector<update_func> update_hook;

/*************************************************************************
	Configurations - Not front-end specific
*************************************************************************/

enum game_mode_type {
	COMPUTER_NONE,
	COMPUTER_BLACK,
	COMPUTER_WHITE,
	COMPUTER_BOTH
};
extern game_mode_type game_mode;

extern const char *game_mode_name[4];

struct level_info_type {
	int midgame_depth;
	int num_empty_winlossdraw;
	int num_empty_exact;
};

const int NUM_LEVEL_INFO = 5;
extern level_info_type level_info[NUM_LEVEL_INFO];

const int LEVEL_CUSTOM = NUM_LEVEL_INFO;
const int LEVEL_CUSTOM_FILE = 99;
extern int computer_level;

extern level_info_type current_level_info;

extern const char *level_name[NUM_LEVEL_INFO+1];

const int NUM_OPENING_VAR = 6;
extern int opening_var;
extern const char *opening_var_name[NUM_OPENING_VAR];

extern bool log_opening;	// Not user configurable yet
extern bool log_move;
extern bool animate_opening;
extern bool redo_ai_move;

const int NUM_ANIMATE_DELAY = 5;
extern int animate_delay;

extern string opening_name;
extern int opening_player;
extern deque<int> opening_move_queue;

const int NUM_START_GAME_MODE = 2;
enum start_game_mode_type {
	START_GAME_INITIAL,
	START_GAME_RANDOM
};
extern const char *start_game_mode_name[NUM_START_GAME_MODE];
extern start_game_mode_type start_game_mode;
extern int start_random_game_pieces;

extern bool show_toolbar;
extern bool show_pattern_evaluation;
extern bool show_game_history;

extern bool show_border;
extern bool hint_move;
extern string theme_name;

/*************************************************************************
	Other functions that have to be provided by front ends
*************************************************************************/

void load_pixmaps(bool);
void resize_board_widget();
void draw_board();
void computer_move();
void cancel_computer_thread();
void new_game();

/*************************************************************************
	Game mode functions
*************************************************************************/

game_mode_type get_game_mode();
void set_game_mode(game_mode_type game_mode_);
const char *get_game_mode_string();
bool is_computer_player(int color);
bool get_wait_player(int color);
bool get_game_play();

/*************************************************************************
	File I/O functions
*************************************************************************/

void user_home_dir_init();
void log_history();

/*************************************************************************
	State update
*************************************************************************/

void update_state(update_state_type);
void add_update_hook(update_func f);
void remove_update_hook(update_func f);

/*************************************************************************
	Utility functions
*************************************************************************/

const char *find_opening_name();
