/*
	grhino.cc	GRhino Gnome Frontend
	Copyright (c) 2000, 2001, 2002, 2003, 2004 Kriang Lerdsuwanakij
	email:		lerdsuwa@users.sourceforge.net

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
#define GNOME_DISABLE_DEPRECATED
#define GTK_DISABLE_DEPRECATED
*/

#include "config.h"

#include <gtk/gtk.h>
#include <gnome.h>

#include <pthread.h>

#include <pwd.h>
#include <sys/types.h>
#include <time.h>

#include CXX__HEADER_exception
#include CXX__HEADER_stdexcept
#include CXX__HEADER_fstream
#include CXX__HEADER_sstream
#include CXX__HEADER_iomanip
#include <string>
#include <deque>
#include <vector>

USING_NAMESPACE_STD;

#include "black.xpm"
#include "white.xpm"
#include "empty.xpm"
#include "hint_black.xpm"
#include "hint_white.xpm"

#include "top.xpm"
#include "bottom.xpm"
#include "left.xpm"
#include "right.xpm"

#include "icon_new.xpm"
#include "icon_empty.xpm"
#include "icon_black.xpm"
#include "icon_white.xpm"

#include "board.h"
#include "hash.h"
#include "alphabeta.h"
#include "opening.h"
#include "pattern.h"
#include "parity.h"
#include "book.h"
#include "game.h"
#include "gpref.h"
#include "randboard.h"

#include "gtstream.h"

// Grab version number in VERSION variable
#undef VERSION
char *
#include "scripts/version"
;

const char *prog_name = "GRhino";
const char *prog_ver = VERSION;

/*************************************************************************
	User interface variables
*************************************************************************/

GdkPixbuf *top_pixmap = 0, *bottom_pixmap = 0, *left_pixmap = 0, *right_pixmap = 0;
GdkPixbuf *black_pixmap = 0, *white_pixmap = 0, *empty_pixmap = 0;
GdkPixbuf *hint_black_pixmap = 0, *hint_white_pixmap = 0;
int grid_width;
int grid_height;
int left_margin;
int top_margin;
int right_margin;
int bottom_margin;
GtkWidget *board_widget;

GtkWidget *status_widget;

GtkWidget *main_window;
GtkWidget *toolbar;
GtkWidget *button_new, *button_exit, *button_undo, *button_redo;
GtkWidget *button_begin, *button_prev, *button_next, *button_end;
int icon_width;
int icon_height;
GtkWidget *icon_new, *icon_exit, *icon_undo, *icon_redo;
GtkWidget *icon_begin, *icon_prev, *icon_next, *icon_end;

board_info *view_board_ptr;
bool view_mode;
int view_position;
int view_player;
GdkCursor *view_cursor;

int gamehist_width, gamehist_height, gamehist_x, gamehist_y;
int pattern_x, pattern_y;
int main_x, main_y;

/*************************************************************************
	View mode
*************************************************************************/

void set_view_mode(bool b)
{
	if (view_cursor == NULL && board_widget)
		view_cursor = gdk_cursor_new_for_display(
				gtk_widget_get_display(board_widget),
				GDK_X_CURSOR);
	if (b)
		gdk_window_set_cursor(board_widget->window, view_cursor);
	else
		gdk_window_set_cursor(board_widget->window, NULL);
	view_mode = b;
}

/*************************************************************************
	Configuration file functions
*************************************************************************/

#define CFG_AI_PATH "GRhino/AI/"
#define CFG_BOARD_PATH "GRhino/Board/"
#define CFG_THEME_PATH "GRhino/Theme/"
#define CFG_VIEW_PATH "GRhino/View/"

void write_view_config()
{
	gnome_config_set_bool(CFG_VIEW_PATH "show_toolbar", show_toolbar);

	gnome_config_set_bool(CFG_VIEW_PATH "show_pattern_evaluation", show_pattern_evaluation);
	gnome_config_set_int(CFG_VIEW_PATH "pattern_evaluation_x", pattern_x);
	gnome_config_set_int(CFG_VIEW_PATH "pattern_evaluation_y", pattern_y);

	gnome_config_set_bool(CFG_VIEW_PATH "show_game_history", show_game_history);
	gnome_config_set_int(CFG_VIEW_PATH "game_history_width", gamehist_width);
	gnome_config_set_int(CFG_VIEW_PATH "game_history_height", gamehist_height);
	gnome_config_set_int(CFG_VIEW_PATH "game_history_x", gamehist_x);
	gnome_config_set_int(CFG_VIEW_PATH "game_history_y", gamehist_y);

	gnome_config_set_int(CFG_VIEW_PATH "main_window_x", main_x);
	gnome_config_set_int(CFG_VIEW_PATH "main_window_y", main_y);

	gnome_config_sync();
}

void write_config()
{
	gnome_config_set_int(CFG_AI_PATH "game_mode", static_cast<int>(get_game_mode()));

	if (computer_level == LEVEL_CUSTOM)
		gnome_config_set_int(CFG_AI_PATH "computer_level", LEVEL_CUSTOM_FILE);
	else
		gnome_config_set_int(CFG_AI_PATH "computer_level", computer_level);
	gnome_config_set_int(CFG_AI_PATH "midgame_depth", current_level_info.midgame_depth);
	gnome_config_set_int(CFG_AI_PATH "num_empty_winlossdraw", current_level_info.num_empty_winlossdraw);
	gnome_config_set_int(CFG_AI_PATH "num_empty_exact", current_level_info.num_empty_exact);

	gnome_config_set_int(CFG_AI_PATH "randomness", get_eval_random());
	gnome_config_set_int(CFG_AI_PATH "opening_var", opening_var);

	gnome_config_set_bool(CFG_BOARD_PATH "log_move", log_move);
	gnome_config_set_bool(CFG_BOARD_PATH "redo_ai_move", redo_ai_move);
	gnome_config_set_bool(CFG_BOARD_PATH "animate_opening", animate_opening);
	gnome_config_set_int(CFG_BOARD_PATH "animate_delay", animate_delay);
	gnome_config_set_int(CFG_BOARD_PATH "start_game_mode", static_cast<int>(start_game_mode));
	gnome_config_set_string(CFG_BOARD_PATH "opening", opening_name.c_str());
	gnome_config_set_int(CFG_BOARD_PATH "start_random_game_pieces", start_random_game_pieces);

						// This is in CFG_BOARD_PATH for
						// compatibility with older GRhino
	gnome_config_set_bool(CFG_BOARD_PATH "hint_move", hint_move);
	gnome_config_set_bool(CFG_THEME_PATH "show_border", show_border);
	gnome_config_set_int(CFG_THEME_PATH "toolbar_icon_size", toolbar_icon_size);
	gnome_config_set_string(CFG_THEME_PATH "theme", theme_name.c_str());

						// Sync in side this call
	write_view_config();
}

void read_config()
{
	gboolean use_def;
						// Default to COMPUTER_WHITE
	int i = gnome_config_get_int_with_default(CFG_AI_PATH "game_mode=2", &use_def);
	if (i < 0 || i >= 4)
		i = COMPUTER_WHITE;
	set_game_mode(static_cast<game_mode_type>(i));

	computer_level = gnome_config_get_int_with_default(CFG_AI_PATH "computer_level=2", &use_def);
	if (computer_level == LEVEL_CUSTOM)
		computer_level = LEVEL_CUSTOM;
	else if (computer_level < 0 || computer_level >= NUM_LEVEL_INFO+1)
		computer_level = 2;

	if (computer_level == LEVEL_CUSTOM) {
		current_level_info.midgame_depth = 
			gnome_config_get_int_with_default(CFG_AI_PATH "midgame_depth=6", &use_def);
		if (current_level_info.midgame_depth < 1 || current_level_info.midgame_depth > 20)
			current_level_info.midgame_depth = 6;

		current_level_info.num_empty_winlossdraw = 
			gnome_config_get_int_with_default(CFG_AI_PATH "num_empty_winlossdraw=12", &use_def);
		if (current_level_info.num_empty_winlossdraw < 1 || current_level_info.num_empty_winlossdraw > 20)
			current_level_info.num_empty_winlossdraw = 12;

		current_level_info.num_empty_exact = 
			gnome_config_get_int_with_default(CFG_AI_PATH "num_empty_exact=10", &use_def);
		if (current_level_info.num_empty_exact < 1 || current_level_info.num_empty_exact > 20)
			current_level_info.num_empty_exact = 10;
	}
	else {
		current_level_info.midgame_depth = level_info[computer_level].midgame_depth;
		current_level_info.num_empty_winlossdraw = level_info[computer_level].num_empty_winlossdraw;
		current_level_info.num_empty_exact = level_info[computer_level].num_empty_exact;
	}

	i = gnome_config_get_int_with_default(CFG_AI_PATH "randomness=0", &use_def);
	if (i < 1 || i > 10)
		i = 0;
	set_eval_random(i);

						// Default to Low
	opening_var = gnome_config_get_int_with_default(CFG_AI_PATH "opening_var=2", &use_def);
	if (opening_var < 0 || opening_var >= NUM_OPENING_VAR)
		opening_var = 1;
	set_book_random(opening_var);

	log_move = gnome_config_get_bool_with_default(CFG_BOARD_PATH "log_move=false", &use_def);
	redo_ai_move = gnome_config_get_bool_with_default(CFG_BOARD_PATH "redo_ai_move=true", &use_def);
	animate_opening = gnome_config_get_bool_with_default(CFG_BOARD_PATH "animate_opening=true", &use_def);

	animate_delay = gnome_config_get_int_with_default(CFG_BOARD_PATH "animate_delay=1", &use_def);
	if (animate_delay < 1 || animate_delay > NUM_ANIMATE_DELAY)
		animate_delay = 1;

	i = gnome_config_get_int_with_default(CFG_BOARD_PATH "start_game_mode=0", &use_def);
	if (i < 0 || i >= NUM_START_GAME_MODE)
		i = 0;
	start_game_mode = static_cast<start_game_mode_type>(i);

	char *str = gnome_config_get_string_with_default(CFG_BOARD_PATH "opening=None", &use_def);
	int id = get_opening_id(str);
	g_free(str);

	i = gnome_config_get_int_with_default(CFG_BOARD_PATH "start_random_game_pieces=20", &use_def);
	if (i < 4 || i > 32 || (i & 1))
		i = 20;
	start_random_game_pieces = i;

						// This is in CFG_BOARD_PATH for
						// compatibility with older GRhino
	hint_move = gnome_config_get_bool_with_default(CFG_BOARD_PATH "hint_move=false", &use_def);
	show_border = gnome_config_get_bool_with_default(CFG_THEME_PATH "show_border=false", &use_def);
	theme_name = gnome_config_get_string_with_default(CFG_THEME_PATH "theme=", &use_def);
	toolbar_icon_size = gnome_config_get_int_with_default(CFG_THEME_PATH "toolbar_icon_size=0", &use_def);
	if (toolbar_icon_size < 0 || toolbar_icon_size >= NUM_TOOLBAR_ICON_SIZE)
		toolbar_icon_size = 0;

	show_toolbar = gnome_config_get_bool_with_default(CFG_VIEW_PATH "show_toolbar=false", &use_def);

	show_pattern_evaluation = gnome_config_get_bool_with_default(CFG_VIEW_PATH "show_pattern_evaluation=false", &use_def);
	pattern_x = gnome_config_get_int_with_default(CFG_VIEW_PATH "pattern_evaluation_x=-1", &use_def);
	pattern_y = gnome_config_get_int_with_default(CFG_VIEW_PATH "pattern_evaluation_y=-1", &use_def);

	show_game_history = gnome_config_get_bool_with_default(CFG_VIEW_PATH "show_game_history=false", &use_def);
	gamehist_width = gnome_config_get_int_with_default(CFG_VIEW_PATH "game_history_width=120", &use_def);
	if (gamehist_width <= 0)
		gamehist_width = 120;
	gamehist_height = gnome_config_get_int_with_default(CFG_VIEW_PATH "game_history_height=300", &use_def);
	if (gamehist_height <= 0)
		gamehist_height = 300;
	gamehist_x = gnome_config_get_int_with_default(CFG_VIEW_PATH "game_history_x=-1", &use_def);
	gamehist_y = gnome_config_get_int_with_default(CFG_VIEW_PATH "game_history_y=-1", &use_def);

	main_x = gnome_config_get_int_with_default(CFG_VIEW_PATH "main_window_x=-1", &use_def);
	main_y = gnome_config_get_int_with_default(CFG_VIEW_PATH "main_window_y=-1", &use_def);

	opening_name = get_opening_name(id);
	opening_player = BLACK;
	get_opening_move_sequence(id, opening_move_queue);
}

/*************************************************************************
	Drawing functions
*************************************************************************/

void draw_board()
{
	for (int x = 0; x < 8; ++x)
		for (int y = 0; y < 8; ++y) {
			GdkPixbuf *pixmap;
			int pos = xy_to_pos(x, y);
			switch (view_board_ptr->board[pos]) {
				case BLACK:
					pixmap = black_pixmap;
					break;
				case WHITE:
					pixmap = white_pixmap;
					break;
				default:
					if (hint_move
					    && can_play(view_board_ptr, player, pos)) {
						if (player == BLACK)
							pixmap = hint_black_pixmap;
						else
							pixmap = hint_white_pixmap;
					}
					else
						pixmap = empty_pixmap;
			}
			gdk_draw_pixbuf(board_widget->window, 
					board_widget->style->fg_gc[GTK_STATE_NORMAL],
					pixmap,
					0, 0, x*grid_width+left_margin, y*grid_height+top_margin,
					grid_width, grid_height,
					GDK_RGB_DITHER_NORMAL, 0, 0);
		}

	if (show_border) {
		gdk_draw_pixbuf(board_widget->window, 
				board_widget->style->fg_gc[GTK_STATE_NORMAL],
				top_pixmap,
				0, 0, 0, 0,
				8*grid_width+left_margin+right_margin, top_margin,
				GDK_RGB_DITHER_NORMAL, 0, 0);
		gdk_draw_pixbuf(board_widget->window, 
				board_widget->style->fg_gc[GTK_STATE_NORMAL],
				bottom_pixmap,
				0, 0, 0, 8*grid_height+top_margin,
				8*grid_width+left_margin+right_margin, bottom_margin,
				GDK_RGB_DITHER_NORMAL, 0, 0);
		gdk_draw_pixbuf(board_widget->window, 
				board_widget->style->fg_gc[GTK_STATE_NORMAL],
				left_pixmap,
				0, 0, 0, top_margin,
				left_margin, 8*grid_height,
				GDK_RGB_DITHER_NORMAL, 0, 0);
		gdk_draw_pixbuf(board_widget->window, 
				board_widget->style->fg_gc[GTK_STATE_NORMAL],
				right_pixmap,
				0, 0, 8*grid_width+left_margin, top_margin,
				right_margin, 8*grid_height,
				GDK_RGB_DITHER_NORMAL, 0, 0);
	}

	gdk_flush();
}

void update_status_bar(update_state_type /*u*/)
{
	ostringstream buf;
	if (game_play) {
		gtout(buf, _("%$ (%$)"))
		     << (player == BLACK ? _("Black's turn") : _("White's turn"))
		     << (wait_player ? _("human") :
				(move_queue.size() ? _("Opening") : _("AI")));

		if (start_game_mode == START_GAME_INITIAL) {
			const char *s = find_opening_name();
			if (s && strcmp(s, "None"))
				gtout(buf, _("   [%$]")) << _(s);
		}
	}
	else {
		int	black_score = board_black_score(board_ptr);
		int	white_score = board_white_score(board_ptr);
		adjust_score(black_score, white_score);
		gtout(buf, _("%$ (%$-%$)"))
		     << (black_score == white_score ? _("Draw") :
			 (black_score > white_score ? _("Black wins") : _("White wins")))
		     << (black_score > white_score ? black_score : white_score)
		     << (black_score > white_score ? white_score : black_score);
	}

	if (view_mode) {
		gtout(buf, _("  (%$ %$)"))
		     << "View move" << view_position;
	}

	buf << ends;

	gnome_appbar_pop(GNOME_APPBAR(status_widget));
	gnome_appbar_push(GNOME_APPBAR(status_widget), buf.str().c_str());
}

void place_piece_and_update(int pos)
{
	bool	during_opening = false;

	if (move_queue.size()) {	// Move come from move queue
		move_queue.pop_front();	// Remove it
		during_opening = true;
	}

	place_piece(board_ptr, player, pos);
	if (can_play(board_ptr, switch_color(player))) {
		player = switch_color(player);
		wait_player = get_wait_player(player);
	}
	else if (!can_play(board_ptr, player)) {
		wait_player = false;
		game_play = false;
	}

	copy_board(view_board_ptr, board_ptr);
	set_view_mode(false);

				// Update history for undo
	bool keep_history = false;
	copy_board(board_history+num_history, board_ptr);
				// Repeat same move
	if (num_history < max_num_history 
	    && move_history[num_history-1] == pos)
		keep_history = true;
	move_history[num_history-1] = pos;
	player_history[num_history] = player;
	num_history++;

				// Remove redo capability
	if (!keep_history)
		max_num_history = num_history;

	if (!during_opening || animate_opening) {
		draw_board();
		update_state(UPDATE_FORWARD);
	}

	if (!game_play)
		log_history();
}

/*************************************************************************
	Computer AI
*************************************************************************/

void *computer_thread(void *)
{
	int last_type;
				// Set deferred cancellation to avoid
				// destroying hash table data structure
	pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, &last_type);

	int pos;
	pthread_mutex_lock(&ai_lock);
	pos = ai_pos;
	pthread_mutex_unlock(&ai_lock);

	if (move_queue.size()) {
				// Remove entry from move_queue in
				// place_piece_and_update since move
				// can be canceled
		pos = move_queue.front();
	}
				// Computer strategy selection
	else if (get_num_move(board_ptr)
		 >= NUM_MOVE - current_level_info.num_empty_exact) {
		eval_endgame(board_ptr, player, &pos);
		free_hash_move(get_num_move(board_ptr));
	}
	else if (get_num_move(board_ptr)
		 >= NUM_MOVE - current_level_info.num_empty_winlossdraw) {
		eval_winlossdraw(board_ptr, player, &pos);
		free_hash_move(get_num_move(board_ptr));
	}
	else if (random_game)
		eval_midgame(board_ptr, player, &pos);
	else {
		int book_pos = book_move(board_ptr, 
					 move_history, 
					 player_history);
		if (book_pos != POS_INVALID) {
			pos = book_pos;
		}
		else
			eval_midgame(board_ptr, player, &pos);
	}

	pthread_mutex_lock(&ai_lock);
	ai_pos = pos;
	pthread_mutex_unlock(&ai_lock);

	return NULL;
}

gint computer_check(void *)
{
	int pos;
	pthread_mutex_lock(&ai_lock);
	pos = ai_pos;

	if (pos != POS_INVALID) {
		gtk_timeout_remove(ai_timeout_id);
		ai_timeout_id = 0;
		void *status;
		pthread_join(ai_thread, &status);

		pthread_mutex_unlock(&ai_lock);

		place_piece_and_update(pos);

		if (game_play == true && wait_player == false)
			computer_move();
	}
	else
		pthread_mutex_unlock(&ai_lock);

	return TRUE;
}

void computer_move()
{
	pthread_mutex_lock(&ai_lock);
	ai_pos = POS_INVALID;
	set_midgame_depth(current_level_info.midgame_depth);
				// Start computer thread
	pthread_create(&ai_thread, &ai_thread_attr, computer_thread, NULL);
				// Repeatly check if computer is finished
	ai_timeout_id = gtk_timeout_add(100 * animate_delay, 
					computer_check, NULL);
	pthread_mutex_unlock(&ai_lock);
}

void cancel_computer_thread()
{
				// Terminate AI thread 
	pthread_mutex_lock(&ai_lock);
	if (ai_timeout_id) {
		pthread_cancel(ai_thread);

		gtk_timeout_remove(ai_timeout_id);
		ai_timeout_id = 0;
		void *status;
		pthread_join(ai_thread, &status);
	}
	pthread_mutex_unlock(&ai_lock);
}

/*************************************************************************
	Game operations
*************************************************************************/

void init_game_before_draw()
{
	read_config();

	user_home_dir_init();
}

void update_menu_and_toolbar(update_state_type);

void init_game()
{
	trans_table_init();
	set_max_hash_move(57);

	pattern_table_init();
	book_init();
	srand(time(NULL));

	pthread_attr_init(&ai_thread_attr);
//	pthread_attr_setdetachstate(&ai_thread_attr, PTHREAD_CREATE_DETACHED);

	add_update_hook(update_status_bar);
	add_update_hook(update_menu_and_toolbar);
}

void new_game()
{
	eval_new_game();

	if (start_game_mode == START_GAME_RANDOM) {
		random_game = true;
		random_board(board_ptr, start_random_game_pieces);
		move_queue.clear();
		player = BLACK;
	}
	else {
		random_game = false;
		copy_board(board_ptr, init_board_ptr);
		move_queue = opening_move_queue;
		player = opening_player;
	}

	copy_board(view_board_ptr, board_ptr);
	set_view_mode(false);

	wait_player = get_wait_player(player);
	game_play = get_game_play();
	draw_board();

				// Update history for undo
	copy_board(board_history+0, board_ptr);
	player_history[0] = player;
	num_history = 1;
	max_num_history = 1;
	min_num_history = opening_move_queue.size() + 1;

				// Make sure history is setup so that
				// Undo/redo capability is properly
				// computed.
	update_state(UPDATE_ALL);

	ai_tampered = false;

	if (game_play == true && wait_player == false) {
		if (move_queue.size() && !animate_opening) {
				// Place all the moves immediately
			do {
				int pos = move_queue.front();
				place_piece_and_update(pos);
			} while (move_queue.size());
				// Draw board only once after all moves
				// are placed
			draw_board();
			update_state(UPDATE_FORWARD);
				// game_play and wait_player are updated
				// place_piece_and_update call above
			if (game_play == true && wait_player == false)
				computer_move();
		}
		else
			computer_move();
	}
}

// Initial board (num_history = 1, min_num_history = 1, max_num_history = 1)
//   board_history   .
//   move_history
//   player_history  B
// After first move F5 (num_history = 2, min_num_history = 1, max_num_history = 2)
//   board_history   .  .
//   move_history   F5
//   player_history  B  W

bool is_undoable()
{
	int	i = num_history;
	if (view_mode)
		return false;
	if (i == min_num_history)	// Already at the beginning of the game
		return false;

	return true;
}

void undo()
{
	cancel_computer_thread();

	int	i = num_history;

				// Skip AI's turns
	while (i > min_num_history && is_computer_player(player_history[i-2]))
		i--;

	if (i > min_num_history)		// Undo human move
		i--;

	copy_board(board_ptr, board_history+i-1);
	copy_board(view_board_ptr, board_ptr);
	set_view_mode(false);

	player = player_history[i-1];
	game_play = get_game_play();
	wait_player = game_play ? get_wait_player(player) : false;

	num_history = i;

	draw_board();
	update_state(UPDATE_BACKWARD);

				// Need to start AI thread
	if (game_play == true && wait_player == false)
		computer_move();
}

bool is_redoable()
{
	int	i = num_history;
	if (view_mode)
		return false;
	if (i == max_num_history)	// Already at the beginning of the game
		return false;

	return true;
}

void redo()
{
	cancel_computer_thread();

	int	i = num_history;

				// Skip AI's turns
	while (i < max_num_history && is_computer_player(player_history[i-1]))
		i++;

	if (i < max_num_history)		// Redo human move
		i++;

	if (redo_ai_move) {
				// Skip AI's turns
		while (i < max_num_history 
		       && is_computer_player(player_history[i-1]))
			i++;
	}

	copy_board(board_ptr, board_history+i-1);
	copy_board(view_board_ptr, board_ptr);
	set_view_mode(false);

	player = player_history[i-1];
	game_play = get_game_play();
	wait_player = game_play ? get_wait_player(player) : false;

	num_history = i;

	draw_board();
	update_state(UPDATE_FORWARD);

				// Need to start AI thread
	if (game_play == true && wait_player == false)
		computer_move();
}

void switch_computer_color()
{
	cancel_computer_thread();

	if (num_history-1 != min_num_history)
		ai_tampered = true;

	switch (get_game_mode()) {
		case COMPUTER_BLACK:
			set_game_mode(COMPUTER_WHITE);
			break;
		case COMPUTER_WHITE:
			set_game_mode(COMPUTER_BLACK);
			break;
		default:
				// COMPUTER_NONE or COMPUTER_BOTH
			break;
	}

	wait_player = game_play ? get_wait_player(player) : false;
	update_state(UPDATE_NONE);

				// Need to start AI thread
	if (game_play == true && wait_player == false)
		computer_move();
}

/*************************************************************************
	Signals
*************************************************************************/

void destroy_signal(GtkWidget *window, gpointer /*data*/)
{
	gtk_window_get_position(GTK_WINDOW(window), &main_x, &main_y);
	write_view_config();
	gtk_main_quit();
}

void expose_signal(GtkWidget * /*window*/, GdkEventExpose * /*event*/, gpointer /*data*/)
{
	draw_board();
}

void button_press_signal(GtkWidget * /*window*/, GdkEventButton *event, gpointer /*data*/)
{
	if (!view_mode
	    && wait_player && event->type == GDK_BUTTON_PRESS
	    && event->button == 1
	    && event->x >= left_margin && event->x < 8*grid_width+left_margin
	    && event->y >= top_margin && event->y < 8*grid_height+top_margin) {

		int x = static_cast<int>((event->x - left_margin)/ grid_width);
		int y = static_cast<int>((event->y - top_margin)/ grid_height);
		int pos = xy_to_pos(x, y);

		if (can_play(board_ptr, player, pos)) {
			place_piece_and_update(pos);
			if (game_play == true && wait_player == false)
				computer_move();
		}
	}
}

/*************************************************************************
	Tools menu - Pattern evaluation
*************************************************************************/

GtkWidget *clist_eval;
GtkWidget *label_score;
GtkWidget *label_move_index;
GtkWidget *label_parity;

void evaluate_position_fill_entry(pattern_t p, int index, int row, int col)
{
	static gchar s[5];	// Enough space for -127 .. 127 and string
				// termination
	int i = 0;

		// pattern_eval returns a signed char (-127 .. 127)
	int score = pattern_eval(view_board_ptr, p, index);
	if (score < 0) {
		s[i++] = '-';
		score = -score;
	}

	int score_left = score;
	if (score >= 100) {
		s[i++] = '1';
		score_left -= 100;
	}
	if (score >= 10) {
		s[i++] = score_left/10 + '0';
		score_left %= 10;
	}
	s[i++] = score_left + '0';
	s[i] = '\0';

	gtk_clist_set_text(GTK_CLIST(clist_eval), row, col, s);
}

void evaluate_position_fill()
{
	static gchar s[10];
	snprintf(s, 10, "%d", static_cast<int>(pattern_eval(view_board_ptr)));
	gtk_label_set_text(GTK_LABEL(label_score), s);
	snprintf(s, 10, "%d", static_cast<int>(to_move_index(view_board_ptr->move)));
	gtk_label_set_text(GTK_LABEL(label_move_index), s);
	snprintf(s, 10, "%d",
		 can_play(view_board_ptr) ? static_cast<int>(parity_eval(view_board_ptr, 
									 view_player))
					    : 0);
	gtk_label_set_text(GTK_LABEL(label_parity), s);

	evaluate_position_fill_entry(PATTERN_ROW1, 0, 0, 2);
	evaluate_position_fill_entry(PATTERN_ROW1, 1, 0, 4);
	evaluate_position_fill_entry(PATTERN_ROW1, 2, 1, 2);
	evaluate_position_fill_entry(PATTERN_ROW1, 3, 1, 4);
	evaluate_position_fill_entry(PATTERN_ROW2, 0, 2, 2);
	evaluate_position_fill_entry(PATTERN_ROW2, 1, 2, 4);
	evaluate_position_fill_entry(PATTERN_ROW2, 2, 3, 2);
	evaluate_position_fill_entry(PATTERN_ROW2, 3, 3, 4);
	evaluate_position_fill_entry(PATTERN_ROW3, 0, 4, 2);
	evaluate_position_fill_entry(PATTERN_ROW3, 1, 4, 4);
	evaluate_position_fill_entry(PATTERN_ROW3, 2, 5, 2);
	evaluate_position_fill_entry(PATTERN_ROW3, 3, 5, 4);
	evaluate_position_fill_entry(PATTERN_ROW4, 0, 6, 2);
	evaluate_position_fill_entry(PATTERN_ROW4, 1, 6, 4);
	evaluate_position_fill_entry(PATTERN_ROW4, 2, 7, 2);
	evaluate_position_fill_entry(PATTERN_ROW4, 3, 7, 4);

	evaluate_position_fill_entry(PATTERN_DIAG1, 0, 8, 2);
	evaluate_position_fill_entry(PATTERN_DIAG1, 1, 8, 4);
	evaluate_position_fill_entry(PATTERN_DIAG2, 0, 9, 2);
	evaluate_position_fill_entry(PATTERN_DIAG2, 1, 9, 4);
	evaluate_position_fill_entry(PATTERN_DIAG2, 2, 10, 2);
	evaluate_position_fill_entry(PATTERN_DIAG2, 3, 10, 4);
	evaluate_position_fill_entry(PATTERN_DIAG3, 0, 11, 2);
	evaluate_position_fill_entry(PATTERN_DIAG3, 1, 11, 4);
	evaluate_position_fill_entry(PATTERN_DIAG3, 2, 12, 2);
	evaluate_position_fill_entry(PATTERN_DIAG3, 3, 12, 4);
	evaluate_position_fill_entry(PATTERN_DIAG4, 0, 13, 2);
	evaluate_position_fill_entry(PATTERN_DIAG4, 1, 13, 4);
	evaluate_position_fill_entry(PATTERN_DIAG4, 2, 14, 2);
	evaluate_position_fill_entry(PATTERN_DIAG4, 3, 14, 4);
	evaluate_position_fill_entry(PATTERN_DIAG5, 0, 15, 2);
	evaluate_position_fill_entry(PATTERN_DIAG5, 1, 15, 4);
	evaluate_position_fill_entry(PATTERN_DIAG5, 2, 16, 2);
	evaluate_position_fill_entry(PATTERN_DIAG5, 3, 16, 4);

	evaluate_position_fill_entry(PATTERN_CORNER5X2, 0, 17, 2);
	evaluate_position_fill_entry(PATTERN_CORNER5X2, 1, 17, 4);
	evaluate_position_fill_entry(PATTERN_CORNER5X2, 2, 18, 2);
	evaluate_position_fill_entry(PATTERN_CORNER5X2, 3, 18, 4);
	evaluate_position_fill_entry(PATTERN_CORNER5X2, 4, 19, 2);
	evaluate_position_fill_entry(PATTERN_CORNER5X2, 5, 19, 4);
	evaluate_position_fill_entry(PATTERN_CORNER5X2, 6, 20, 2);
	evaluate_position_fill_entry(PATTERN_CORNER5X2, 7, 20, 4);
}

void evaluate_position_update(update_state_type /*u*/)
{
	gtk_clist_freeze(GTK_CLIST(clist_eval));
	evaluate_position_fill();
	gtk_clist_thaw(GTK_CLIST(clist_eval));
}

gint evaluate_position_configure_event(GtkWidget *widget, 
				       GdkEventConfigure * /*e*/,
				       gpointer /*data*/)
{
	gtk_window_get_position(GTK_WINDOW(widget), &pattern_x, &pattern_y);
	return FALSE;		// Propagate signal to child so that it
				// can resize itself
}

void evaluate_position_closed(GtkWidget *widget, GtkWidget **data);

void evaluate_position()
{
	static gchar *empty = "";
	static gchar *row[5];

	static GtkWidget *evalbox;
	if (evalbox) {
		if (show_pattern_evaluation)
			gtk_window_present(GTK_WINDOW(evalbox));
		else
			evaluate_position_closed(evalbox, &evalbox);
		return;
	}

	if (!show_pattern_evaluation)
		return;

	row[0] = _("Pattern name");
	row[1] = _("Position");
	row[2] = _("Score");
	row[3] = _("Position");
	row[4] = _("Score");

	evalbox = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	if (pattern_x != -1 && pattern_y != -1)
		gtk_window_move(GTK_WINDOW(evalbox), pattern_x, pattern_y);
	gtk_window_set_title(GTK_WINDOW(evalbox), _("Pattern Evaluation"));
	gtk_container_set_border_width(GTK_CONTAINER(evalbox), 5);
	g_signal_connect(G_OBJECT(evalbox), "destroy",
			 G_CALLBACK(evaluate_position_closed), &evalbox);
	g_signal_connect(G_OBJECT(evalbox), "configure_event",
			 G_CALLBACK(evaluate_position_configure_event), NULL);

	GtkWidget *vbox = gtk_vbox_new(FALSE, 5);

	GtkWidget *hbox1 = gtk_hbox_new(TRUE, 5);
	GtkWidget *label1 = gtk_label_new(_("Total score"));
	gtk_box_pack_start(GTK_BOX(hbox1), label1, TRUE, TRUE, 2);
	label_score = gtk_label_new("");
	gtk_box_pack_start(GTK_BOX(hbox1), label_score, TRUE, TRUE, 2);
	gtk_box_pack_start(GTK_BOX(vbox), hbox1, TRUE, TRUE, 2);

	GtkWidget *hbox2 = gtk_hbox_new(TRUE, 5);
	GtkWidget *label2 = gtk_label_new(_("Move index"));
	gtk_box_pack_start(GTK_BOX(hbox2), label2, TRUE, TRUE, 2);
	label_move_index = gtk_label_new("");
	gtk_box_pack_start(GTK_BOX(hbox2), label_move_index, TRUE, TRUE, 2);
	gtk_box_pack_start(GTK_BOX(vbox), hbox2, TRUE, TRUE, 2);

	GtkWidget *hbox3 = gtk_hbox_new(TRUE, 5);
	GtkWidget *label3 = gtk_label_new(_("Parity score"));
	gtk_box_pack_start(GTK_BOX(hbox3), label3, TRUE, TRUE, 2);
	label_parity = gtk_label_new("");
	gtk_box_pack_start(GTK_BOX(hbox3), label_parity, TRUE, TRUE, 2);
	gtk_box_pack_start(GTK_BOX(vbox), hbox3, TRUE, TRUE, 2);

	clist_eval = gtk_clist_new_with_titles(5, row);
	gtk_box_pack_start(GTK_BOX(vbox), clist_eval, TRUE, TRUE, 2);
	gtk_container_add(GTK_CONTAINER(evalbox), vbox);

	gtk_clist_column_titles_passive(GTK_CLIST(clist_eval));
	row[2] = empty;
	row[4] = empty;
	row[0] = _("Row 1");
	row[1] = _("A1-H1");
	row[3] = _("A8-H8");
	gtk_clist_append(GTK_CLIST(clist_eval), row);
	row[0] = empty;
	row[1] = _("A1-A8");
	row[3] = _("H1-H8");
	gtk_clist_append(GTK_CLIST(clist_eval), row);
	row[0] = _("Row 2");
	row[1] = _("A2-H2");
	row[3] = _("A7-H7");
	gtk_clist_append(GTK_CLIST(clist_eval), row);
	row[0] = empty;
	row[1] = _("B1-B8");
	row[3] = _("G1-G8");
	gtk_clist_append(GTK_CLIST(clist_eval), row);
	row[0] = _("Row 3");
	row[1] = _("A3-H3");
	row[3] = _("A6-H6");
	gtk_clist_append(GTK_CLIST(clist_eval), row);
	row[0] = empty;
	row[1] = _("C1-C8");
	row[3] = _("F1-F8");
	gtk_clist_append(GTK_CLIST(clist_eval), row);
	row[0] = _("Row 4");
	row[1] = _("A4-H4");
	row[3] = _("A5-H5");
	gtk_clist_append(GTK_CLIST(clist_eval), row);
	row[0] = empty;
	row[1] = _("D1-D8");
	row[3] = _("E1-E8");
	gtk_clist_append(GTK_CLIST(clist_eval), row);

	row[0] = _("Diag 1");
	row[1] = _("A1-H8");
	row[3] = _("H1-A8");
	gtk_clist_append(GTK_CLIST(clist_eval), row);
	row[0] = _("Diag 2");
	row[1] = _("A2-G8");
	row[3] = _("B1-H7");
	gtk_clist_append(GTK_CLIST(clist_eval), row);
	row[0] = empty;
	row[1] = _("H2-B8");
	row[3] = _("G1-A7");
	gtk_clist_append(GTK_CLIST(clist_eval), row);
	row[0] = _("Diag 3");
	row[1] = _("A3-F8");
	row[3] = _("C1-H6");
	gtk_clist_append(GTK_CLIST(clist_eval), row);
	row[0] = empty;
	row[1] = _("H3-C8");
	row[3] = _("F1-A6");
	gtk_clist_append(GTK_CLIST(clist_eval), row);
	row[0] = _("Diag 4");
	row[1] = _("A4-E8");
	row[3] = _("D1-H5");
	gtk_clist_append(GTK_CLIST(clist_eval), row);
	row[0] = empty;
	row[1] = _("H4-D8");
	row[3] = _("E1-A5");
	gtk_clist_append(GTK_CLIST(clist_eval), row);
	row[0] = _("Diag 5");
	row[1] = _("A5-D8");
	row[3] = _("E1-H4");
	gtk_clist_append(GTK_CLIST(clist_eval), row);
	row[0] = empty;
	row[1] = _("H5-E8");
	row[3] = _("D1-A4");
	gtk_clist_append(GTK_CLIST(clist_eval), row);

	row[0] = _("Corner 5x2");
	row[1] = _("A1-E2");
	row[3] = _("H1-D2");
	gtk_clist_append(GTK_CLIST(clist_eval), row);
	row[0] = empty;
	row[1] = _("A8-E7");
	row[3] = _("H8-D7");
	gtk_clist_append(GTK_CLIST(clist_eval), row);
	row[1] = _("A1-B5");
	row[3] = _("A8-B4");
	gtk_clist_append(GTK_CLIST(clist_eval), row);
	row[1] = _("H1-G5");
	row[3] = _("H8-G4");
	gtk_clist_append(GTK_CLIST(clist_eval), row);

	evaluate_position_fill();
	add_update_hook(evaluate_position_update);

	gtk_widget_show_all(evalbox);
}

/*************************************************************************
	Tools menu - Game history
*************************************************************************/

GtkWidget *tree_view_history;
GtkListStore *list_store_history;
int list_store_size;

GdkPixbuf *scale_pixbuf_from_xpm_data(const char *xpm[], int width, int height)
{
	GdkPixbuf *p1, *p2;
	p1 = gdk_pixbuf_new_from_xpm_data(xpm);
	if (!p1)
		return p1;
	p2 = gdk_pixbuf_scale_simple(p1, width, height, GDK_INTERP_BILINEAR);
	g_object_unref(p1);
	return p2;
}

void game_history_fill()
{
	static gchar m[NUM_MOVE][3];	// Enough space for A1 .. H8 and string
				// termination
	static gchar p[NUM_MOVE][3];
	static gchar *pos_table_x;
	static gchar *pos_table_y;
	static GdkPixbuf *color_table[3];
	static GdkPixbuf *empty_player_pixbuf;   
	static GdkPixbuf *black_player_pixbuf;
	static GdkPixbuf *white_player_pixbuf;

	static bool init = false;
	if (!init) {
				// I don't think we need i18n string for these
		pos_table_x = "ABCDEFGH";
		pos_table_y = "12345678";
		for (int i = 0; i < NUM_MOVE; ++i) {
				// String termination
			m[i][2] = '\0';
			if (i+1 < 10)
				p[i][0] = ' ';
			else
				p[i][0] = (i+1)/10+'0';
			p[i][1] = (i+1)%10+'0';
			p[i][2] = '\0';
		}

		int player_size = 12;
		empty_player_pixbuf = scale_pixbuf_from_xpm_data(const_cast<const char**>(icon_empty_xpm),
								 player_size, player_size);
		black_player_pixbuf = scale_pixbuf_from_xpm_data(const_cast<const char**>(icon_black_xpm),
								 player_size, player_size);
		white_player_pixbuf = scale_pixbuf_from_xpm_data(const_cast<const char**>(icon_white_xpm),
								 player_size, player_size);

		color_table[0] = white_player_pixbuf;
		color_table[2] = black_player_pixbuf;

		init = true;
	}

	GtkTreeIter iter;

	if (list_store_size == 0) {
		gtk_list_store_append(list_store_history, &iter);
		gtk_list_store_set(list_store_history, &iter, 
				   0, 0, 
				   1, "    ", 
				   2, empty_player_pixbuf, 
				   3, _("Begin"), -1);
		++list_store_size;
	}

	if (list_store_size < num_history)
		for (int i = list_store_size; i < num_history; ++i) {
			int pos = move_history[i-1];
			m[i-1][0] = pos_table_x[pos_to_x(pos)];
			m[i-1][1] = pos_table_y[pos_to_y(pos)];
						// WHITE (-1) => 0
						// BLACK (1) => 2
			GdkPixbuf *c = color_table[player_history[i-1]+1];

			gtk_list_store_append(list_store_history, &iter);
			gtk_list_store_set(list_store_history, &iter, 
					   0, i, 
					   1, p[i-1], 
					   2, c, 
					   3, m[i-1], -1);
			++list_store_size;
		}
	else if (list_store_size > num_history) {
		GtkTreePath *path = gtk_tree_path_new();
		gtk_tree_path_append_index(path, num_history);
		gtk_tree_model_get_iter(GTK_TREE_MODEL(list_store_history), 
					&iter, path);
		gtk_tree_path_free(path);
		while (list_store_size > num_history) {
			gtk_list_store_remove(list_store_history, &iter);
			--list_store_size;
		}
	}
}

void game_history_update(update_state_type u)
{
	switch (u) {
		case UPDATE_NONE:
			break;		// Just update selection
		case UPDATE_BACKWARD:
		case UPDATE_FORWARD:
			game_history_fill();
			break;
		case UPDATE_ALL:
			gtk_list_store_clear(list_store_history);
			list_store_size = 0;
			game_history_fill();
			break;
	}

	GtkTreePath *path = gtk_tree_path_new();
	if (view_mode)
		gtk_tree_path_append_index(path, view_position);
	else
		gtk_tree_path_append_index(path, num_history-1);

	GtkTreeSelection *selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree_view_history));
	gtk_tree_selection_select_path(selection, path);
	gtk_tree_view_scroll_to_cell(GTK_TREE_VIEW(tree_view_history), path, 
				     NULL, FALSE, 0, 0);
	gtk_tree_path_free(path);
// Scroll problem when scrollbar appears
}

gint game_history_configure_event(GtkWidget *widget, 
				  GdkEventConfigure *e, gpointer /*data*/)
{
	gtk_window_get_position(GTK_WINDOW(widget), &gamehist_x, &gamehist_y);
	gamehist_width = e->width;
	gamehist_height = e->height;
	return FALSE;		// Propagate signal to child so that it
				// can resize itself
}

void game_history_closed(GtkWidget *widget, GtkWidget **data);

void game_history_selected(GtkTreeSelection *selection, gpointer /*data*/)
{
	GtkTreeIter iter;
	GtkTreeModel *model;
	if (gtk_tree_selection_get_selected(selection, &model, &iter)) {
		int pos;
		gtk_tree_model_get(model, &iter, 0, &pos, -1);
		if (pos != num_history-1) {
			set_view_mode(true);
			view_position = pos;
			view_player = player_history[view_position];
			copy_board(view_board_ptr, board_history+view_position);
			draw_board();
			update_state(UPDATE_NONE);
		}
		else if (view_mode == true) { // pos == num_history-1
			set_view_mode(false);
			copy_board(view_board_ptr, board_ptr);
			draw_board();
			update_state(UPDATE_NONE);
		}
	}
}

void game_history()
{
	static GtkWidget *gamehist = NULL;
	if (gamehist) {
		if (show_game_history)
			gtk_window_present(GTK_WINDOW(gamehist));
		else
			game_history_closed(gamehist, &gamehist);
		return;
	}

	if (!show_game_history)
		return;

	gamehist = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(gamehist), _("Game History"));
	gtk_container_set_border_width(GTK_CONTAINER(gamehist), 5);
	g_signal_connect(G_OBJECT(gamehist), "destroy",
			 G_CALLBACK(game_history_closed), &gamehist);
	g_signal_connect(G_OBJECT(gamehist), "configure_event",
			 G_CALLBACK(game_history_configure_event), NULL);

				// Reuse old list_store_history
	if (list_store_history) {
		gtk_list_store_clear(list_store_history);
	}
	else {
		list_store_history = gtk_list_store_new(
						4, G_TYPE_INT, G_TYPE_STRING,
						G_TYPE_OBJECT, G_TYPE_STRING);
	}
	list_store_size = 0;

	GtkWidget *scroll = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll),
				       GTK_POLICY_NEVER,
				       GTK_POLICY_AUTOMATIC);
	if (gamehist_x != -1 && gamehist_y != -1)
		gtk_window_move(GTK_WINDOW(gamehist), gamehist_x, gamehist_y);
	gtk_window_set_default_size(GTK_WINDOW(gamehist), gamehist_width, gamehist_height);
	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scroll),
					    GTK_SHADOW_ETCHED_IN);
	gtk_container_add(GTK_CONTAINER(gamehist), scroll);

	tree_view_history = gtk_tree_view_new_with_model(GTK_TREE_MODEL(list_store_history));
	GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
	g_object_set(G_OBJECT(renderer), "xalign", 1.0, NULL);	// Right aligned
	GtkTreeViewColumn *column = gtk_tree_view_column_new_with_attributes(
						"Number",
						renderer,
						"text", 1,
						NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(tree_view_history),
				    column);
	column = gtk_tree_view_column_new_with_attributes(
						"Color",
						gtk_cell_renderer_pixbuf_new(),
						"pixbuf", 2,
						NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(tree_view_history),
				    column);
	column = gtk_tree_view_column_new_with_attributes(
						"Move",
						gtk_cell_renderer_text_new(),
						"text", 3,
						NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(tree_view_history),
				    column);

	gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(tree_view_history),
					  FALSE);
// This causes lots of flicker
	gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(tree_view_history), TRUE);

	GtkTreeSelection *selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree_view_history));
	gtk_tree_selection_set_mode (selection, GTK_SELECTION_SINGLE);
	g_signal_connect(G_OBJECT(selection), "changed",
			 G_CALLBACK(game_history_selected), NULL);

	gtk_container_add(GTK_CONTAINER(scroll), tree_view_history);

	game_history_fill();
	add_update_hook(game_history_update);

	gtk_widget_show_all(gamehist);

	GtkTreePath *path = gtk_tree_path_new();
	gtk_tree_path_append_index(path, num_history-1);
	gtk_tree_selection_select_path(selection, path);
	gtk_tree_view_scroll_to_cell(GTK_TREE_VIEW(tree_view_history), path, 
				     NULL, FALSE, 0, 0);
	gtk_tree_path_free(path);
}

/*************************************************************************
	Adjust toolbar according to current settings
*************************************************************************/

bool toolbar_docked = false;

void update_toolbar()
{
	if (toolbar_docked) {
		BonoboDockItem *dock_item = gnome_app_get_dock_item_by_name(GNOME_APP(main_window),
									    GNOME_APP_TOOLBAR_NAME);
		if (show_toolbar)
			gtk_widget_show(GTK_WIDGET(dock_item));
		else {
			gtk_widget_hide(GTK_WIDGET(dock_item));
			gtk_window_resize(GTK_WINDOW(main_window), 1, 1);
		}
	}
	else if (show_toolbar) {
		gnome_app_set_toolbar(GNOME_APP(main_window), 
				      GTK_TOOLBAR(toolbar));
		toolbar_docked = true;
		update_toolbar();
	}
}

/*************************************************************************
	Menu callback functions
*************************************************************************/

void menu_game_new(GtkWidget * /*widget*/, gpointer /*data*/)
{
	new_game();
}

void menu_game_undo(GtkWidget * /*widget*/, gpointer /*data*/)
{
	undo();
}

void menu_game_redo(GtkWidget * /*widget*/, gpointer /*data*/)
{
	redo();
}

void menu_game_exit(GtkWidget * /*widget*/, gpointer /*data*/)
{
	gtk_main_quit();
}

void menu_game_begin(GtkWidget * /*widget*/, gpointer /*data*/)
{
	copy_board(view_board_ptr, board_history+0);
	view_player = player_history[0];
	set_view_mode(true);
	view_position = 0;
	draw_board();
	update_state(UPDATE_NONE);
}

void menu_game_prev(GtkWidget * /*widget*/, gpointer /*data*/)
{
	if (!view_mode) {
		set_view_mode(true);
		view_position = num_history-2;
	}
	else
		--view_position;
	copy_board(view_board_ptr, board_history+view_position);
	view_player = player_history[view_position];
	draw_board();
	update_state(UPDATE_NONE);
}

void menu_game_next(GtkWidget * /*widget*/, gpointer /*data*/)
{
	++view_position;
	if (view_position == num_history-1)
		set_view_mode(false);
	copy_board(view_board_ptr, board_history+view_position);
	view_player = player_history[view_position];
	draw_board();
	update_state(UPDATE_NONE);
}

void menu_game_end(GtkWidget * /*widget*/, gpointer /*data*/)
{
	copy_board(view_board_ptr, board_ptr);
	set_view_mode(false);
	draw_board();
	update_state(UPDATE_NONE);
}

void menu_view_toolbar(GtkWidget * /*widget*/, gpointer /*data*/);
void menu_tools_evaluate_position(GtkWidget * /*widget*/, gpointer /*data*/);
void menu_tools_game_history(GtkWidget * /*widget*/, gpointer /*data*/);

void menu_settings_save(GtkWidget * /*widget*/, gpointer /*data*/)
{
	write_config();
}

void menu_settings_preferences(GtkWidget * /*widget*/, gpointer /*data*/)
{
	preferences();
}

void menu_settings_switch_color(GtkWidget * /*widget*/, gpointer /*data*/)
{
	switch_computer_color();
}

void menu_help_about(GtkWidget * /*widget*/, gpointer /*data*/)
{
	static GtkWidget *about_dialog = NULL;
	if (!about_dialog) {
		static const gchar *authors[2];
		authors[0] = _("Kriang Lerdsuwanakij");
		authors[1] = 0;
		static const gchar *documenters[2];
		documenters[0] = _("Kriang Lerdsuwanakij");
		documenters[1] = 0;
		gchar *translator = _("TRANSLATOR");
		if (strcmp (translator, "TRANSLATOR") == 0)
			translator = 0;

		about_dialog = gnome_about_new(_(prog_name), prog_ver,
					       _("(c) 2000-3, 2004 Kriang Lerdsuwanakij"),
					       _("An Othello/Reversi program"),
					       authors,
					       documenters,
					       translator,
					       0);
		g_signal_connect(G_OBJECT(about_dialog), "destroy",
				 G_CALLBACK(gtk_widget_destroyed),
				 &about_dialog);
	}
	gtk_window_present(GTK_WINDOW(about_dialog));
}

/*************************************************************************
	Menu Info
*************************************************************************/

GnomeUIInfo menu_game_info[] = {
	GNOMEUIINFO_MENU_NEW_GAME_ITEM(menu_game_new, 0),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_MENU_QUIT_ITEM(menu_game_exit, 0),
	GNOMEUIINFO_END
};

GnomeUIInfo menu_edit_info[] = {
	GNOMEUIINFO_MENU_UNDO_MOVE_ITEM(menu_game_undo, 0),
	GNOMEUIINFO_MENU_REDO_MOVE_ITEM(menu_game_redo, 0),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_ITEM_STOCK(N_("_Begin of game"), N_("Go to the game beginning"), menu_game_begin, GTK_STOCK_GOTO_TOP),
	GNOMEUIINFO_ITEM_STOCK(N_("_Previous move"), N_("Go to the previous move"), menu_game_prev, GTK_STOCK_GO_UP),
	GNOMEUIINFO_ITEM_STOCK(N_("_Next move"), N_("Go to the next move"), menu_game_next, GTK_STOCK_GO_DOWN),
	GNOMEUIINFO_ITEM_STOCK(N_("_Last move"), N_("Go to the last move"), menu_game_end, GTK_STOCK_GOTO_BOTTOM),
	GNOMEUIINFO_END
};

GnomeUIInfo menu_tools_info[] = {
	GNOMEUIINFO_TOGGLEITEM(N_("_Toolbar"), N_("Show/hide toolbar"), menu_view_toolbar, NULL),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_TOGGLEITEM(N_("_Pattern evaluation"), N_("Show/hide pattern evaluation score"), menu_tools_evaluate_position, NULL),
	GNOMEUIINFO_TOGGLEITEM(N_("Game _history"), N_("Show/hide game move history"), menu_tools_game_history, NULL),
	GNOMEUIINFO_END
};

GnomeUIInfo menu_settings_info[] = {
	GNOMEUIINFO_ITEM_NONE(N_("Switch _computer color"), NULL, menu_settings_switch_color),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_ITEM_NONE(N_("_Save Settings"), NULL, menu_settings_save),
	GNOMEUIINFO_MENU_PREFERENCES_ITEM(menu_settings_preferences, 0),
	GNOMEUIINFO_END
};

GnomeUIInfo menu_help_info[] = {
	GNOMEUIINFO_HELP("grhino"),
	GNOMEUIINFO_MENU_ABOUT_ITEM(menu_help_about, 0),
	GNOMEUIINFO_END
};

GnomeUIInfo menu_main_info[] = {
	GNOMEUIINFO_MENU_GAME_TREE(menu_game_info),
	GNOMEUIINFO_MENU_EDIT_TREE(menu_edit_info),
	GNOMEUIINFO_SUBTREE(N_("_View"), menu_tools_info),
	GNOMEUIINFO_MENU_SETTINGS_TREE(menu_settings_info),
	GNOMEUIINFO_MENU_HELP_TREE(menu_help_info),
	GNOMEUIINFO_END
};

/*************************************************************************
	Pixmap function
*************************************************************************/

void default_pixmap_error(const char *str)
{
	gtstream bufstr;
	gtout(bufstr, _("size of theme pixmap %$ does not agree with others"))
		<< str;
	throw range_error(bufstr.str());
}

GdkPixbuf *load_pixmap(const char *basename, const char *default_xpm[], 
		       bool use_default)
{
	if (use_default)
		return gdk_pixbuf_new_from_xpm_data(default_xpm);
	else {
		string pathname = THEMEDIR;
		pathname.append(theme_name);
		pathname.append("/");
		pathname.append(basename);
		string filename = pathname;
		filename.append(".png");
		if (!g_file_test(filename.c_str(), G_FILE_TEST_EXISTS)) {
			filename = pathname;
			filename.append(".xpm");
			if (!g_file_test(filename.c_str(), G_FILE_TEST_EXISTS)) {
				gtstream bufstr;
				gtout(bufstr, _("pixmap %$ missing"))
					<< basename;
				throw runtime_error(bufstr.str());
			}
		}
		GdkPixbuf *pixmap = gdk_pixbuf_new_from_file(filename.c_str(), 
							     NULL);
		if (!pixmap) {
			gtstream bufstr;
			gtout(bufstr, _("pixmap %$ contains error"))
				<< filename;
			throw runtime_error(bufstr.str());
		}
		return pixmap;
	}
}

void load_icon_image(GtkWidget *widget, const char *basename, 
		     const char *stock_name, 
		     bool use_default)
{
	if (use_default) {
		gtk_image_set_from_stock(GTK_IMAGE(widget), stock_name, 
					 toolbar_icon_map[toolbar_icon_size]);
		return;
	}
	else {
		string pathname = THEMEDIR;
		pathname.append(theme_name);
		pathname.append("/");
		pathname.append(basename);
		string filename = pathname;
		filename.append(".png");
		if (!g_file_test(filename.c_str(), G_FILE_TEST_EXISTS)) {
			filename = pathname;
			filename.append(".xpm");
			if (!g_file_test(filename.c_str(), G_FILE_TEST_EXISTS)) {
				gtk_image_set_from_stock(GTK_IMAGE(widget), 
							 stock_name, 
							 toolbar_icon_map[toolbar_icon_size]);
				return;
			}
		}
		GdkPixbuf *pixbuf = gdk_pixbuf_new_from_file(filename.c_str(), 
							     NULL);
		if (!pixbuf) {
			gtk_image_set_from_stock(GTK_IMAGE(widget), 
						 stock_name, 
						 toolbar_icon_map[toolbar_icon_size]);
			return;
		}
		GdkPixbuf *pixbuf2 = gdk_pixbuf_scale_simple(pixbuf, 
							     icon_width, 
							     icon_height, 
							     GDK_INTERP_BILINEAR);
		gtk_image_set_from_pixbuf(GTK_IMAGE(widget), pixbuf2);
		g_object_unref(pixbuf);
		g_object_unref(pixbuf2);
		return;
	}
}

void load_icon_image(GtkWidget *widget, const char *basename, 
		     const char *default_xpm[], 
		     bool use_default)
{
	if (use_default) {
		GdkPixbuf *p = scale_pixbuf_from_xpm_data(default_xpm, 
							  icon_width, 
							  icon_height);
		gtk_image_set_from_pixbuf(GTK_IMAGE(widget), p);
		g_object_unref(p);
		return;
	}
	else {
		string pathname = THEMEDIR;
		pathname.append(theme_name);
		pathname.append("/");
		pathname.append(basename);
		string filename = pathname;
		filename.append(".png");
		if (!g_file_test(filename.c_str(), G_FILE_TEST_EXISTS)) {
			filename = pathname;
			filename.append(".xpm");
			if (!g_file_test(filename.c_str(), G_FILE_TEST_EXISTS)) {
				GdkPixbuf *p = scale_pixbuf_from_xpm_data(default_xpm, 
									  icon_width, 
									  icon_height);
				gtk_image_set_from_pixbuf(GTK_IMAGE(widget), p);
				g_object_unref(p);
				return;
			}
		}
		GdkPixbuf *pixbuf = gdk_pixbuf_new_from_file(filename.c_str(), NULL);
		if (!pixbuf) {
			GdkPixbuf *p = scale_pixbuf_from_xpm_data(default_xpm, 
								  icon_width, 
								  icon_height);
			gtk_image_set_from_pixbuf(GTK_IMAGE(widget), p);
			g_object_unref(p);
			return;
		}
		GdkPixbuf *pixbuf2 = gdk_pixbuf_scale_simple(pixbuf, icon_width, 
							     icon_height, 
							     GDK_INTERP_BILINEAR);
		gtk_image_set_from_pixbuf(GTK_IMAGE(widget), pixbuf2);
		g_object_unref(pixbuf);
		g_object_unref(pixbuf2);
		return;
	}
}

void load_pixmaps(bool use_default)
{
					// Free memory used by old pixmaps
	if (black_pixmap) {
		g_object_unref(black_pixmap);
		black_pixmap = 0;
	}
	if (white_pixmap) {
		g_object_unref(white_pixmap);
		white_pixmap = 0;
	}
	if (empty_pixmap) {
		g_object_unref(empty_pixmap);
		empty_pixmap = 0;
	}
	if (hint_black_pixmap) {
		g_object_unref(hint_black_pixmap);
		hint_black_pixmap = 0;
	}
	if (hint_white_pixmap) {
		g_object_unref(hint_white_pixmap);
		hint_white_pixmap = 0;
	}
	if (left_pixmap) {
		g_object_unref(left_pixmap);
		left_pixmap = 0;
	}
	if (right_pixmap) {
		g_object_unref(right_pixmap);
		right_pixmap = 0;
	}
	if (top_pixmap) {
		g_object_unref(top_pixmap);
		top_pixmap = 0;
	}
	if (bottom_pixmap) {
		g_object_unref(bottom_pixmap);
		bottom_pixmap = 0;
	}

	black_pixmap = load_pixmap("black", const_cast<const char**>(black_xpm), 
				   use_default);
	grid_width = gdk_pixbuf_get_width(black_pixmap);
	grid_height = gdk_pixbuf_get_height(black_pixmap);

	white_pixmap = load_pixmap("white", const_cast<const char**>(white_xpm), 
				   use_default);
	if (gdk_pixbuf_get_width(white_pixmap) != grid_width
	    || gdk_pixbuf_get_height(white_pixmap) != grid_height)
		default_pixmap_error("white");
	
	empty_pixmap = load_pixmap("empty", const_cast<const char**>(empty_xpm), 
				   use_default);
	if (gdk_pixbuf_get_width(empty_pixmap) != grid_width
	    || gdk_pixbuf_get_height(empty_pixmap) != grid_height)
		default_pixmap_error("empty");

	hint_black_pixmap = load_pixmap("hint_black", const_cast<const char**>(hint_black_xpm), 
					use_default);
	if (gdk_pixbuf_get_width(hint_black_pixmap) != grid_width
	    || gdk_pixbuf_get_height(hint_black_pixmap) != grid_height)
		default_pixmap_error("hint_black");

	hint_white_pixmap = load_pixmap("hint_white", const_cast<const char**>(hint_white_xpm), 
					use_default);
	if (gdk_pixbuf_get_width(hint_white_pixmap) != grid_width
	    || gdk_pixbuf_get_height(hint_white_pixmap) != grid_height)
		default_pixmap_error("hint_white");

				// Setup border pixmaps if required
	if (show_border) {
		left_pixmap = load_pixmap("left", const_cast<const char**>(left_xpm), 
					  use_default);
		left_margin = gdk_pixbuf_get_width(left_pixmap);
		if (gdk_pixbuf_get_height(left_pixmap) != grid_height * 8)
			default_pixmap_error("left");

		right_pixmap = load_pixmap("right", const_cast<const char**>(right_xpm), 
					   use_default);
		right_margin = gdk_pixbuf_get_width(right_pixmap);
		if (gdk_pixbuf_get_height(right_pixmap) != grid_height * 8)
			default_pixmap_error("right");

		top_pixmap = load_pixmap("top", const_cast<const char**>(top_xpm), 
					 use_default);
		top_margin = gdk_pixbuf_get_height(top_pixmap);
		if (gdk_pixbuf_get_width(top_pixmap) != grid_width * 8 + left_margin + right_margin)
			default_pixmap_error("top");

		bottom_pixmap = load_pixmap("bottom", const_cast<const char**>(bottom_xpm), 
					    use_default);
		bottom_margin = gdk_pixbuf_get_height(bottom_pixmap);
		if (gdk_pixbuf_get_width(bottom_pixmap) != grid_width * 8 + left_margin + right_margin)
			default_pixmap_error("bottom");
	}
	else {
		left_margin = 0;
		right_margin = 0;
		top_margin = 0;
		bottom_margin = 0;
	}

	GdkScreen *screen = gtk_window_get_screen(GTK_WINDOW(main_window));
	GtkSettings *settings = gtk_settings_get_for_screen(screen);
	gtk_icon_size_lookup_for_settings(settings, toolbar_icon_map[toolbar_icon_size], 
					  &icon_width, &icon_height);

	load_icon_image(icon_new, "icon_new", const_cast<const char**>(icon_new_xpm), 
			use_default);
	load_icon_image(icon_exit, "icon_exit", GTK_STOCK_QUIT, use_default);
	load_icon_image(icon_undo, "icon_undo", GTK_STOCK_UNDO, use_default);
	load_icon_image(icon_redo, "icon_redo", GTK_STOCK_REDO, use_default);

	load_icon_image(icon_begin, "icon_begin", GTK_STOCK_GOTO_TOP, use_default);
	load_icon_image(icon_prev, "icon_prev", GTK_STOCK_GO_UP, use_default);
	load_icon_image(icon_next, "icon_next", GTK_STOCK_GO_DOWN, use_default);
	load_icon_image(icon_end, "icon_end", GTK_STOCK_GOTO_BOTTOM, use_default);
}

void init_pixmaps()
{
	icon_new = gtk_image_new();
	icon_exit = gtk_image_new();
	icon_undo = gtk_image_new();
	icon_redo = gtk_image_new();

	icon_begin = gtk_image_new();
	icon_prev = gtk_image_new();
	icon_next = gtk_image_new();
	icon_end = gtk_image_new();

	if (theme_name.size()) {
		try {
			load_pixmaps(false);
		}
		catch (exception &e) {
			cout << flush;
			gtout(cerr, _("%$: %$\n")) << prog_name << e.what();
			gtout(cerr, _("%$: reverting to default internal theme\n"))
			     << prog_name;
			load_pixmaps(true);
		}
	}
	else
		load_pixmaps(true);
}

void resize_board_widget()
{
				// Set board size
	gtk_widget_set_size_request(board_widget,
				    8 * grid_width + left_margin + right_margin,
				    8 * grid_height + top_margin + bottom_margin);

				// Set main window size
				// Setting to 1x1 automatically set
				// window size to smallest possible
	gtk_window_resize(GTK_WINDOW(main_window), 1, 1);
}

/*************************************************************************
	Toolbar function
*************************************************************************/

void create_toolbar()
{
	toolbar = gtk_toolbar_new();
	button_new = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
					     _("New"),
					     _("Start a new game"),
					     NULL, icon_new,
					     G_CALLBACK (menu_game_new), 
					     NULL);
	button_exit = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
					      _("Quit"),
					      _("Quit GRhino"),
					      NULL, icon_exit,
					      G_CALLBACK (menu_game_exit), 
					      NULL);
	button_undo = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
					      _("Undo"),
					      _("Undo move"),
					      NULL, icon_undo,
					      G_CALLBACK (menu_game_undo), 
					      NULL);
	button_redo = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
					      _("Redo"),
					      _("Redo move"),
					      NULL, icon_redo,
					      G_CALLBACK (menu_game_redo), 
					      NULL);
	gtk_toolbar_append_space(GTK_TOOLBAR(toolbar));

	button_begin = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
					     _("Begin"),
					     _("Go to the game beginning"),
					     NULL, icon_begin,
					     G_CALLBACK (menu_game_begin), 
					     NULL);
	button_prev = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
					      _("Prev"),
					      _("Go to the previous move"),
					      NULL, icon_prev,
					      G_CALLBACK (menu_game_prev), 
					      NULL);
	button_next = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
					      _("Next"),
					      _("Go to the next move"),
					      NULL, icon_next,
					      G_CALLBACK (menu_game_next),
					      NULL);
	button_end = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
					     _("Last"),
					     _("Go to the last move"),
					     NULL, icon_end,
					     G_CALLBACK (menu_game_end), 
					     NULL);
}

void menu_view_toolbar(GtkWidget * /*widget*/, gpointer /*data*/)
{
	show_toolbar = GTK_CHECK_MENU_ITEM (menu_tools_info[0].widget)->active;
	update_toolbar();
}

void menu_tools_evaluate_position(GtkWidget * /*widget*/, gpointer /*data*/)
{
	show_pattern_evaluation = GTK_CHECK_MENU_ITEM (menu_tools_info[2].widget)->active;
	evaluate_position();
}

void menu_tools_game_history(GtkWidget * /*widget*/, gpointer /*data*/)
{
	show_game_history = GTK_CHECK_MENU_ITEM (menu_tools_info[3].widget)->active;
	game_history();
}

void update_menu_and_toolbar(update_state_type /*u*/)
{
	bool b = is_undoable();
	gtk_widget_set_sensitive(menu_edit_info[0].widget, b);
	gtk_widget_set_sensitive(button_undo, b);

	b = is_redoable();
	gtk_widget_set_sensitive(menu_edit_info[1].widget, b);
	gtk_widget_set_sensitive(button_redo, b);

	b = (!view_mode && (num_history > 1))
	    || (view_mode && (view_position > 0));
	gtk_widget_set_sensitive(button_begin, b);
	gtk_widget_set_sensitive(menu_edit_info[3].widget, b);
	gtk_widget_set_sensitive(button_prev, b);
	gtk_widget_set_sensitive(menu_edit_info[4].widget, b);

	b = view_mode;
	gtk_widget_set_sensitive(button_next, b);
	gtk_widget_set_sensitive(menu_edit_info[5].widget, b);
	gtk_widget_set_sensitive(button_end, b);
	gtk_widget_set_sensitive(menu_edit_info[6].widget, b);
}

// Called from Preferences dialog
void update_menu_and_toolbar_pref()
{
	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM (menu_tools_info[0].widget),
				       show_toolbar);

	game_mode_type g = get_game_mode();
	if (g == COMPUTER_BLACK || g == COMPUTER_WHITE)
		gtk_widget_set_sensitive(menu_settings_info[0].widget, 1);
	else			// COMPUTER_NONE or COMPUTER_BOTH
		gtk_widget_set_sensitive(menu_settings_info[0].widget, 0);

	gtk_toolbar_set_icon_size(GTK_TOOLBAR(toolbar), 
				  toolbar_icon_map[toolbar_icon_size]);
}

void evaluate_position_closed(GtkWidget *widget, GtkWidget **data)
{
	remove_update_hook(evaluate_position_update);
	gtk_widget_destroy(widget);
	*data = NULL;
	show_pattern_evaluation = false;
	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM (menu_tools_info[2].widget),
				       show_pattern_evaluation);
}

void game_history_closed(GtkWidget *widget, GtkWidget **data)
{
	remove_update_hook(game_history_update);
	gtk_widget_destroy(widget);
	*data = NULL;
	show_game_history = false;
	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM (menu_tools_info[3].widget),
				       show_game_history);
}

/*************************************************************************
	Main function
*************************************************************************/

int main_real(gint argc, gchar *argv[])
{
	if (sizeof(int) < 4)
		throw out_of_range(_("cannot run on system with less than 32-bit int"));

	setlocale(LC_ALL, "");
	bindtextdomain("grhino", LOCALEDIR);
	textdomain("grhino");

	gnome_program_init("grhino", prog_ver, LIBGNOMEUI_MODULE, argc, argv,
			   GNOME_PROGRAM_STANDARD_PROPERTIES,
			   NULL);

	init_game_before_draw();

	main_window = gnome_app_new("GRhino", _(prog_name));
	g_signal_connect(G_OBJECT(main_window), "destroy",
			 G_CALLBACK(destroy_signal), 0);
	gtk_widget_realize(main_window);

	gnome_app_create_menus(GNOME_APP(main_window), menu_main_info);
	init_pixmaps();

	board_widget = gtk_drawing_area_new();
				// Set board size
	gtk_widget_set_size_request(board_widget,
				    8 * grid_width + left_margin + right_margin,
				    8 * grid_height + top_margin + bottom_margin);

	g_signal_connect(G_OBJECT(board_widget), "expose_event",
			 G_CALLBACK(expose_signal), 0);
	g_signal_connect(G_OBJECT(board_widget), "button_press_event",
			 G_CALLBACK(button_press_signal), 0);
	gtk_widget_set_events(board_widget,
			      GDK_BUTTON_PRESS_MASK | GDK_EXPOSURE_MASK);
	gtk_widget_show(board_widget);

	board_ptr = new board_info();
	view_board_ptr = new board_info();
	init_board_ptr = new board_info(&board_begin);

	GtkWidget *vbox = gtk_vbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), board_widget, TRUE, TRUE, 0);
	gtk_widget_show(vbox);

	gnome_app_set_contents(GNOME_APP(main_window), vbox);

	status_widget = gnome_appbar_new(FALSE, TRUE, GNOME_PREFERENCES_USER);
	gnome_app_set_statusbar(GNOME_APP(main_window), status_widget);

	create_toolbar();
	if (show_toolbar) {
			// Only dock toolbar when it is visible.
			// This is to avoid broken window size calculation when
			// dock toolbar and immediately make it invisible.
		gnome_app_set_toolbar(GNOME_APP(main_window), GTK_TOOLBAR(toolbar));
		toolbar_docked = true;
	}
	update_menu_and_toolbar_pref();

	if (main_x != -1 && main_y != -1)
		gtk_window_move(GTK_WINDOW(main_window), main_x, main_y);
	gtk_widget_show(main_window);

	init_game();
	new_game();

			// Show other windows after game has
			// been properly setup
	if (show_pattern_evaluation)
		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM (menu_tools_info[2].widget),
					       show_pattern_evaluation);
	if (show_game_history)
		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM (menu_tools_info[3].widget),
					       show_game_history);
	gtk_window_present(GTK_WINDOW(main_window));

	gtk_main();

	return 0;
}

int main(gint argc, gchar *argv[])
{
	try {
		return main_real(argc, argv);
	}
	catch (exception &e) {
		cout << flush;
		gtout(cerr, _("%$: %$\n")) << prog_name << e.what();
	}
	catch (const char *s) {
		cout << flush;
		gtout(cerr, _("%$: exception %$\n")) << prog_name << s;
	}
	catch (...) {
		cout << flush;
		gtout(cerr, _("%$: unknown exception\n")) << prog_name;
	}
	return 1;
}
